import base64

from typing import Self

from cryptography.hazmat.primitives.serialization import KeySerializationEncryption 
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat, PrivateFormat, NoEncryption
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey  

from nacl.bindings import (
    crypto_sign_ed25519_sk_to_curve25519,
    crypto_sign_ed25519_pk_to_curve25519,
)

class XKeyPair:
    def __init__(self, private_key: X25519PrivateKey):
        self._private_key = private_key
        self._public_key = self._private_key.public_key()

    @classmethod
    def generate(cls) -> Self:
        return cls(X25519PrivateKey.generate())

    @classmethod
    def import_from_base64(cls, private_b64: str) -> Self:
        return cls(cls.base64_to_private_key(private_b64))

    @staticmethod
    def base64_to_private_key(b64: str) -> X25519PrivateKey:
        raw_bytes = base64.b64decode(b64)
        return X25519PrivateKey.from_private_bytes(raw_bytes)

    @staticmethod
    def base64_to_public_key(b64: str) -> X25519PublicKey:
        raw_bytes = base64.b64decode(b64)
        return X25519PublicKey.from_public_bytes(raw_bytes)

    @staticmethod
    def private_key_to_bytes(
                private_key: X25519PrivateKey,
                encryption: KeySerializationEncryption = NoEncryption(), 
            ) -> bytes :
        private_bytes = private_key.private_bytes(
            encoding=Encoding.Raw,
            format=PrivateFormat.Raw,
            encryption_algorithm=encryption
        )
        return private_bytes

    @staticmethod
    def public_key_to_bytes(
                public_key: X25519PublicKey,
            ) -> bytes :
        public_bytes = public_key.public_bytes(
            encoding=Encoding.Raw,
            format=PublicFormat.Raw,
        )
        return public_bytes

    @staticmethod
    def private_key_to_base64(
                private_key: X25519PrivateKey,
                encryption: KeySerializationEncryption = NoEncryption(), 
                encoding="utf-8"
            ) -> bytes | str:
        private_bytes = XKeyPair.private_key_to_bytes(private_key, encryption)
        b64 = base64.b64encode(private_bytes)

        if (encoding):
            return b64.decode(encoding)
        return b64

    @staticmethod
    def public_key_to_base64(
                public_key: X25519PublicKey,
                encoding="utf-8"
            ) -> bytes | str:
        public_bytes = XKeyPair.public_key_to_bytes(public_key) 
        b64 = base64.b64encode(public_bytes)

        if (encoding):
            return b64.decode(encoding)
        return b64


    def get_private_key(self) -> X25519PrivateKey:
        return self._private_key

    def get_public_key(self) -> X25519PublicKey:
        return self._public_key

    def get_private_key_bytes(self, encryption=NoEncryption()) -> bytes:
        return XKeyPair.private_key_to_bytes(self._private_key, encryption)

    def get_public_key_bytes(self) -> bytes:
        return XKeyPair.public_key_to_bytes(self._public_key)

    def get_base64_private_key(
                self, 
                encryption: KeySerializationEncryption = NoEncryption(), 
                encoding="utf-8"
            ) -> bytes | str:
        b64 = XKeyPair.private_key_to_base64(
                private_key=self._private_key,
                encryption=encryption,
                encoding=encoding
        )
        return b64 
        
    def get_base64_public_key(
                self, 
                encoding="utf-8"
            ) -> bytes | str:
        b64 = XKeyPair.public_key_to_base64(
                public_key=self._public_key,
                encoding=encoding
        )
        return b64 


class EdKeyPair:
    def __init__(self, private_key: Ed25519PrivateKey):
        self._private_key = private_key
        self._public_key = self._private_key.public_key()

    @classmethod
    def generate(cls) -> Self:
        return cls(Ed25519PrivateKey.generate())

    @classmethod
    def import_from_base64(cls, private_b64: str) -> Self:
        return cls(cls.base64_to_private_key(private_b64))

    @staticmethod
    def base64_to_private_key(b64: str) -> Ed25519PrivateKey:
        raw_bytes = base64.b64decode(b64)
        return Ed25519PrivateKey.from_private_bytes(raw_bytes)

    @staticmethod
    def base64_to_public_key(b64: str) -> Ed25519PublicKey:
        raw_bytes = base64.b64decode(b64)
        return Ed25519PublicKey.from_public_bytes(raw_bytes)

    @staticmethod
    def private_key_to_bytes(
                private_key: Ed25519PrivateKey,
                encryption: KeySerializationEncryption = NoEncryption(), 
            ) -> bytes :
        private_bytes = private_key.private_bytes(
            encoding=Encoding.Raw,
            format=PrivateFormat.Raw,
            encryption_algorithm=encryption
        )
        return private_bytes

    @staticmethod
    def public_key_to_bytes(
                public_key: Ed25519PublicKey,
            ) -> bytes :
        public_bytes = public_key.public_bytes(
            encoding=Encoding.Raw,
            format=PublicFormat.Raw,
        )
        return public_bytes

    @staticmethod
    def private_key_to_base64(
                private_key: Ed25519PrivateKey,
                encryption: KeySerializationEncryption = NoEncryption(), 
                encoding="utf-8"
            ) -> bytes | str:
        private_bytes = EdKeyPair.private_key_to_bytes(private_key, encryption) 
        b64 = base64.b64encode(private_bytes)

        if (encoding):
            return b64.decode(encoding)
        return b64

    @staticmethod
    def public_key_to_base64(
                public_key: Ed25519PublicKey,
                encoding="utf-8"
            ) -> bytes | str:
        public_bytes = EdKeyPair.public_key_to_bytes(public_key) 
        b64 = base64.b64encode(public_bytes)

        if (encoding):
            return b64.decode(encoding)
        return b64

    @staticmethod
    def private_ed_to_x_key(private_key: Ed25519PrivateKey) -> X25519PrivateKey:
        private_bytes = private_key.private_bytes(
            encoding=Encoding.Raw,
            format=PrivateFormat.Raw,
            encryption_algorithm=NoEncryption()
        )
        public_bytes = private_key.public_key().public_bytes(
            encoding=Encoding.Raw,
            format=PublicFormat.Raw,
        )
        ed_bytes = private_bytes + public_bytes
        x_bytes = crypto_sign_ed25519_sk_to_curve25519(ed_bytes)
        return X25519PrivateKey.from_private_bytes(x_bytes)


    @staticmethod
    def public_ed_to_x_key(public_key: Ed25519PublicKey) -> X25519PublicKey:
        pyblic_bytes = public_key.public_bytes(
            encoding=Encoding.Raw,
            format=PublicFormat.Raw,
        )
        x_bytes = crypto_sign_ed25519_pk_to_curve25519(pyblic_bytes)
        return X25519PublicKey.from_public_bytes(x_bytes)


    def get_private_key(self) -> Ed25519PrivateKey:
        return self._private_key

    def get_public_key(self) -> Ed25519PublicKey:
        return self._public_key

    def get_private_key_bytes(self, encryption=NoEncryption()) -> bytes:
        return EdKeyPair.private_key_to_bytes(self._private_key, encryption) 

    def get_public_key_bytes(self) -> bytes:
        return EdKeyPair.public_key_to_bytes(self._public_key) 

    def get_x_private_key(self) -> X25519PrivateKey:
        return EdKeyPair.private_ed_to_x_key(self._private_key)

    def get_x_public_key(self) -> X25519PublicKey:
        return EdKeyPair.public_ed_to_x_key(self._public_key)

    def get_base64_private_key(
                self, 
                encryption: KeySerializationEncryption = NoEncryption(), 
                encoding="utf-8"
            ) -> bytes | str:
        b64 = EdKeyPair.private_key_to_base64(
                private_key=self._private_key,
                encryption=encryption,
                encoding=encoding
        )
        return b64 
        
    def get_base64_public_key(
                self, 
                encoding="utf-8"
            ) -> bytes | str:
        b64 = EdKeyPair.public_key_to_base64(
                public_key=self._public_key,
                encoding=encoding
        )
        return b64 

    def sign_public_key(
                self, 
                public_key: X25519PublicKey | Ed25519PublicKey, 
                encoding="utf-8"
            ) -> bytes | str:
        sign_bytes = self._private_key.sign(public_key.public_bytes(Encoding.Raw, PublicFormat.Raw))

        if (encoding):
            return base64.b64encode(sign_bytes).decode(encoding)
        return sign_bytes 


    @staticmethod
    def verify_public_key(
                verifier: Ed25519PublicKey, 
                public_key: X25519PublicKey | Ed25519PublicKey, 
                signature: bytes,
            ):
        verifier.verify(signature, public_key.public_bytes(Encoding.Raw, PublicFormat.Raw))




import unittest

import base64

from osmomemo import Omemo
from osmomemo import OmemoBundle
from osmomemo import XKeyPair
from osmomemo import EdKeyPair
from osmomemo.storage import OmemoStorage

class TestOmemo(unittest.TestCase):
    def test_init(self):
        storage_path = "./tests/omemo.db"
        storage = OmemoStorage(storage_path)

        bundle_a = OmemoBundle(
            676767676,
            EdKeyPair.generate(),
            XKeyPair.generate(),
            {
                "0": XKeyPair.generate(),
                "1": XKeyPair.generate(),
                "2": XKeyPair.generate(),
            }
        )

        bundle_b = OmemoBundle(
            45454545,
            EdKeyPair.generate(),
            XKeyPair.generate(),
            {
                "0": XKeyPair.generate(),
                "1": XKeyPair.generate(),
                "2": XKeyPair.generate(),
            }
        )

        omemo_a = Omemo(bundle_a, storage)
        omemo_b = Omemo(bundle_b, storage)

        message = "Initial OMEMO message (1234567890)."
        jid_b="bob@domain.com"
        device_b=45454545
        ik_b = bundle_b.get_indentity().get_public_key() 
        spk_b = bundle_b.get_prekey().get_public_key() 
        sign_b = bundle_b.get_prekey_signature(encoding=None) 
        opk_id = "0"
        opk_b = bundle_b.get_onetime_prekey(opk_id).get_public_key()
        

        EK_A, en_message = omemo_a.create_init_message(
            jid=jid_b,
            device=device_b,
            message_bytes=message.encode(),
            indentity_key=ik_b,
            signed_prekey=spk_b,
            prekey_signature=sign_b,
            onetime_prekey=opk_b,
        )

        jid_a="alice@domain.com"
        device_a=676767676
        ik_a = bundle_a.get_indentity().get_public_key() 
        ek_a = EK_A
        spk_id = "0"
        msg = en_message

        de_message = omemo_b.accept_init_message(
            jid=jid_a,
            device=device_a,
            encrypted_message=msg,
            indentity_key=ik_a,
            ephemeral_key=ek_a,
            spk_id=spk_id,
            opk_id=opk_id
        )

        self.assertEqual(message, de_message.decode())

    def test_send(self):
        storage_path = "./tests/omemo.db"
        storage = OmemoStorage(storage_path)

        bundle_a = OmemoBundle(
            676767676,
            EdKeyPair.generate(),
            XKeyPair.generate(),
            {
                "0": XKeyPair.generate(),
                "1": XKeyPair.generate(),
                "2": XKeyPair.generate(),
            }
        )

        bundle_b = OmemoBundle(
            45454545,
            EdKeyPair.generate(),
            XKeyPair.generate(),
            {
                "0": XKeyPair.generate(),
                "1": XKeyPair.generate(),
                "2": XKeyPair.generate(),
            }
        )

        omemo_a = Omemo(bundle_a, storage)
        omemo_b = Omemo(bundle_b, storage)

        message = "Initial OMEMO message (1234567890)."
        jid_b="bob@domain.com"
        device_b=bundle_b.get_device_id()
        ik_b = bundle_b.get_indentity().get_public_key() 
        spk_b = bundle_b.get_prekey().get_public_key() 
        sign_b = bundle_b.get_prekey_signature(encoding=None) 
        opk_id = "0"
        opk_b = bundle_b.get_onetime_prekey(opk_id).get_public_key()
        

        EK_A, en_message = omemo_a.create_init_message(
            jid=jid_b,
            device=device_b,
            message_bytes=message.encode(),
            indentity_key=ik_b,
            signed_prekey=spk_b,
            prekey_signature=sign_b,
            onetime_prekey=opk_b,
        )

        jid_a="alice@domain.com"
        device_a=bundle_a.get_device_id()
        ik_a = bundle_a.get_indentity().get_public_key() 
        ek_a = EK_A
        spk_id = "0"
        msg = en_message

        de_message = omemo_b.accept_init_message(
            jid=jid_a,
            device=device_a,
            encrypted_message=msg,
            indentity_key=ik_a,
            ephemeral_key=ek_a,
            spk_id=spk_id,
            opk_id=opk_id
        )

        self.assertEqual(message, de_message.decode())


        ### Sending
        ## Alice
        message_a = "Hello Bob! How are you?"

        wrapped_a, payload_a = omemo_a.send_message(jid_b, device_b, message_a.encode())

        ## Bob
        message_b = omemo_b.receive_message(jid_a, device_a, wrapped_a, payload_a)

        # TEST
        self.assertEqual(message_a, message_b.decode())


        message_b = "Hi, Alice! I am good =)"

        wrapped_b, payload_b = omemo_b.send_message(jid_a, device_a, message_b.encode())

        message_a = omemo_a.receive_message(jid_b, device_b, wrapped_b, payload_b)

        # Test
        self.assertEqual(message_b, message_a.decode())


    def test_signature(self):
        bundle = OmemoBundle(
            34059834,
            EdKeyPair.generate(),
            XKeyPair.generate(),
            {
                "0": XKeyPair.generate(),
                "1": XKeyPair.generate(),
                "2": XKeyPair.generate(),
            }
        )


        ik_b64 = bundle.get_indentity().get_base64_public_key()
        spk_b64 = bundle.get_prekey().get_base64_public_key()
        spk_sign_b64 = bundle.get_prekey_signature()

        EdKeyPair.verify_public_key(
            EdKeyPair.base64_to_public_key(ik_b64),
            XKeyPair.base64_to_public_key(spk_b64),
            base64.b64decode(spk_sign_b64.encode("utf-8")),
        )



import time

from typing import List

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, declarative_base
from sqlalchemy import select

from .models import Base 
from .models import Jid 
from .models import Device 
from .models import Session as OmemoSession 

class OmemoStorage:
    def __init__(self, store_path: str):
        self._engine = create_engine(f"sqlite:///{store_path}")
        Base.metadata.create_all(self._engine)

    def add_device(self, jid: str, device: int) -> None:
        with Session(self._engine) as session:
            stmt = select(Jid).where(Jid.jid == jid)
            ojid = session.scalar(stmt)

            if (ojid):
                ojid.devices.append(Device(device=device))
                session.add(ojid)
            else:
                ojid = Jid(jid=jid, devices=[Device(device=device)])
                session.add(ojid)
            session.commit()

    def add_session(
                self,
                jid: str,
                device: int, 
                receive_secret_key: str, 
                send_secret_key: str, 
                receive_nonce: str, 
                send_nonce: str, 
            ) -> None:
        with Session(self._engine) as session:
            stmt = select(Device).join(Jid).where(Jid.jid == jid).where(Device.device == device)
            odevice = session.scalar(stmt)
            if (odevice):
                osession = OmemoSession(
                    timestamp=time.time(),
                    receive_secret_key=receive_secret_key,
                    send_secret_key=send_secret_key,
                    receive_nonce=receive_nonce,
                    send_nonce=send_nonce,
                )
                odevice.session = osession
                session.commit()
            else:
                raise Exception("Thas JID or device does not exist in the database.")


    def get_device_list(self, jid: str) -> List[int]:
        devices = []
        with Session(self._engine) as session:
            stmt = select(Jid).where(Jid.jid == jid)
            ojid = session.scalar(stmt)

            if (ojid):
                for device in ojid.devices:
                    devices.append(device.device)
            else:
                raise Exception("That JID does not exist in the database.")
        return devices

    def get_session(self, jid: str, device: int) -> OmemoSession:
        with Session(self._engine) as session:
            stmt = select(Device).join(Jid).where(Jid.jid == jid).where(Device.device == device)
            odevice = session.scalar(stmt)
            if (odevice):
                return odevice.session
            else:
                raise Exception("Thas JID or device does not exist in the database.")

    def set_receive_nonce(self, jid: str, device: int, nonce: str) -> None:
        with Session(self._engine) as session:
            stmt = select(Device).join(Jid).where(Jid.jid == jid).where(Device.device == device)
            odevice = session.scalar(stmt)
            if (odevice):
                if (odevice.session):
                    odevice.session.receive_nonce = nonce 
                    session.commit()
                else:
                    raise Exception("No session for this device.")
            else:
                raise Exception("Thas JID or device does not exist in the database.")

    def set_send_nonce(self, jid: str, device: int, nonce: str) -> None:
        with Session(self._engine) as session:
            stmt = select(Device).join(Jid).where(Jid.jid == jid).where(Device.device == device)
            odevice = session.scalar(stmt)
            if (odevice):
                if (odevice.session):
                    odevice.session.send_nence = nonce 
                    session.commit()
                else:
                    raise Exception("No session for this device.")
            else:
                raise Exception("Thas JID or device does not exist in the database.")

    def update_receive_secret(self, jid: str, device: int, secret: str):
        with Session(self._engine) as session:
            stmt = select(Device).join(Jid).where(Jid.jid == jid).where(Device.device == device)
            odevice = session.scalar(stmt)
            if (odevice):
                if (odevice.session):
                    odevice.session.receive_secret_key = secret
                    session.commit()
                else:
                    raise Exception("No session for this device.")
            else:
                raise Exception("Thas JID or device does not exist in the database.")

    def update_send_secret(self, jid: str, device: int, secret: str):
        with Session(self._engine) as session:
            stmt = select(Device).join(Jid).where(Jid.jid == jid).where(Device.device == device)
            odevice = session.scalar(stmt)
            if (odevice):
                if (odevice.session):
                    odevice.session.send_secret_key = secret
                    session.commit()
                else:
                    raise Exception("No session for this device.")
            else:
                raise Exception("Thas JID or device does not exist in the database.")



import unittest

from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey  

from osmomemo import XKeyPair
from osmomemo import EdKeyPair

class TestKeys(unittest.TestCase):
    def test_generating(self):
        xpair = XKeyPair.generate()
        edpair = EdKeyPair.generate()

        self.assertIsInstance(xpair, XKeyPair)
        self.assertIsInstance(edpair, EdKeyPair)

        self.assertIsInstance(xpair.get_private_key(), X25519PrivateKey)
        self.assertIsInstance(xpair.get_public_key(), X25519PublicKey)
        self.assertIsInstance(edpair.get_private_key(), Ed25519PrivateKey)
        self.assertIsInstance(edpair.get_public_key(), Ed25519PublicKey)


    def test_converting(self):
        edpair = EdKeyPair.generate()

        self.assertIsInstance(edpair.get_x_private_key(), X25519PrivateKey)
        self.assertIsInstance(edpair.get_x_public_key(), X25519PublicKey)


        xpair = XKeyPair(edpair.get_x_private_key())

        self.assertEqual(xpair.get_public_key(), edpair.get_x_public_key())


    def test_base64(self):
        xpair = XKeyPair.generate()
        edpair = EdKeyPair.generate()

        x_priv = xpair.get_base64_private_key()
        x_pub = xpair.get_base64_public_key()
        ed_priv = edpair.get_base64_private_key()
        ed_pub = edpair.get_base64_public_key()

        self.assertIsInstance(x_priv, str)
        self.assertIsInstance(x_pub, str)
        self.assertIsInstance(ed_priv, str)
        self.assertIsInstance(ed_pub, str)

        i_xpair = XKeyPair.import_from_base64(x_priv)
        i_edpair = EdKeyPair.import_from_base64(ed_priv)

        self.assertEqual(i_xpair.get_base64_private_key(), x_priv)
        self.assertEqual(i_xpair.get_base64_public_key(), x_pub)
        self.assertEqual(i_edpair.get_base64_private_key(), ed_priv)
        self.assertEqual(i_edpair.get_base64_public_key(), ed_pub)


        # The private ones are not the same for some reason. 
        # But if convert them to bytes, they match.
        #self.assertEqual(i_xpair.get_private_key(), xpair.get_private_key())
        self.assertEqual(i_xpair.get_public_key(), xpair.get_public_key())
        #self.assertEqual(i_edpair.get_private_key(), edpair.get_private_key())
        self.assertEqual(i_edpair.get_public_key(), edpair.get_public_key())



import hashlib

from typing import Dict

from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey  

from .key import XKeyPair, EdKeyPair 

class OmemoBundle:
    def __init__(
                self,
                device_id: int,
                indentity_key: EdKeyPair,
                signed_prekey: XKeyPair,
                onetime_prekeys: Dict[str, XKeyPair],
            ):
        self._device_id = device_id
        self._indentity_key = indentity_key
        self._signed_prekey = signed_prekey
        self._onetime_prekeys = onetime_prekeys

    def get_device_id(self):
        return self._device_id

    def get_indentity(self) -> EdKeyPair:
        return self._indentity_key

    def get_prekey(self) -> XKeyPair:
        return self._signed_prekey

    def get_onetime_prekeys(self) -> Dict[str, XKeyPair]:
        return self._onetime_prekeys
    
    def get_onetime_prekey(self, id: str) -> XKeyPair:
        try:
            return self._onetime_prekeys[id]
        except KeyError:
            raise Exception("There is no PreKey with this ID.")

    def get_prekey_signature(self, encoding="utf-8") -> bytes | str:
        return self._indentity_key.sign_public_key(
                public_key=self._signed_prekey.get_public_key(),
                encoding=encoding
        )
    
    def get_indentity_fingerprint(self) -> bytes:
        public_bytes = self._indentity_key.get_public_key_bytes()
        return OmemoBundle.public_key_to_fingerprint(public_bytes)  

    def get_indentity_hex_fingerprint(self) -> str:
        public_bytes = self._indentity_key.get_public_key_bytes()
        return OmemoBundle.public_key_to_hex_fingerprint(public_bytes)


    @staticmethod
    def public_key_to_fingerprint(public_key_bytes: bytes):
        hash_obj = hashlib.sha256()
        hash_obj.update(public_key_bytes)
        return hash_obj.digest() 

    @staticmethod
    def public_key_to_hex_fingerprint(public_key_bytes: bytes):
        return OmemoBundle.public_key_to_fingerprint(public_key_bytes).hex()




from typing import List
from typing import Optional
from sqlalchemy import ForeignKey
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship

class Base(DeclarativeBase):
    pass

class Jid(Base):
    __tablename__ = "jid"
    id: Mapped[int] = mapped_column(primary_key=True)
    jid: Mapped[str] = mapped_column()
    devices: Mapped[List["Device"]] = relationship(
        back_populates="jid", cascade="all, delete-orphan"
    )
    def __repr__(self) -> str:
        return f"Jid(id={self.id!r}, jid={self.jid!r}"

class Device(Base):
    __tablename__ = "device"
    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int] = mapped_column(ForeignKey("jid.id"))
    device: Mapped[int] = mapped_column()

    jid: Mapped["Jid"] = relationship(back_populates="devices")
    session: Mapped["Session"] = relationship(
        back_populates="device", cascade="all, delete-orphan", uselist=False
    )

    def __repr__(self) -> str:
        return f"Device(id={self.id!r}, device={self.device!r})"


class Session(Base):
    __tablename__ = "session"
    id: Mapped[int] = mapped_column(primary_key=True)
    device_id: Mapped[int] = mapped_column(ForeignKey("device.id"))
    timestamp: Mapped[int] = mapped_column()
    receive_secret_key: Mapped[str] = mapped_column()
    send_secret_key: Mapped[str] = mapped_column()
    receive_nonce: Mapped[str] = mapped_column()
    send_nonce: Mapped[str] = mapped_column()

    device: Mapped["Device"] = relationship(back_populates="session")

    def __repr__(self) -> str:
        return f"Session(id={self.id!r}, timestamp={self.timestamp!r})"



__version__ = "1.1.2"
__author__ = "osmiumnet"

from .omemo import Omemo
from .bundle import OmemoBundle
from .key import XKeyPair, EdKeyPair 

__all__ = [
    "Omemo",
    "OmemoBundle",
    "XKeyPair",
    "EdKeyPair",
]



import os
import json
import base64

from typing import Tuple, List

from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey  

from .bundle import OmemoBundle
from .storage import OmemoStorage
from .key import XKeyPair, EdKeyPair 
from .crypto import OmemoCryptography as OmemoCrypto

def b64(b: bytes) -> str:
    return base64.b64encode(b).decode('utf-8')

def ub64(s: str) -> bytes:
    return base64.b64decode(s.encode("utf-8"))

class Omemo:
    def __init__(self, bundle: OmemoBundle, storage: OmemoStorage):
        self._bundle = bundle
        self._storage = storage

    def get_device_list(self, jid) -> List[int] | None:
        try:
            return self._storage.get_device_list(jid)
        except:
            return None

    def create_init_message(
                self,
                jid: str,
                device: int,
                message_bytes: bytes,
                indentity_key: Ed25519PublicKey,
                signed_prekey: X25519PublicKey,
                prekey_signature: bytes,
                onetime_prekey: X25519PublicKey,
            ) -> Tuple[bytes, bytes, bytes]:
        # Key pairs
        indentity_pair = self._bundle.get_indentity()

        # Create nonce for receive and send
        # Session nonces will be dinamic in future 
        session_nonces = os.urandom(24)

        # Combine nonce with message to encrypt
        message = session_nonces + message_bytes

        SK, ek_pub, encrypted_message = OmemoCrypto.create_init_message(
                message_bytes=message,
                indentity_pair=indentity_pair,
                indentity_key=indentity_key,
                signed_prekey=signed_prekey,
                prekey_signature=prekey_signature,
                onetime_prekey=onetime_prekey,
        )

        SK_RECV, SK_SEND = OmemoCrypto.split_secret_key(SK)

        self._storage.add_device(jid, device)
        self._storage.add_session(
                jid=jid, 
                device=device, 
                receive_secret_key=b64(SK_RECV), 
                send_secret_key=b64(SK_SEND),
                receive_nonce=b64(session_nonces[:12]),
                send_nonce=b64(session_nonces[12:]), 
        )

        return ek_pub, encrypted_message 

    def accept_init_message(
                self,
                jid: str,
                device: int,
                encrypted_message: bytes,
                indentity_key: Ed25519PublicKey,
                ephemeral_key: X25519PublicKey,
                spk_id: str,
                opk_id: str,
            ) -> Tuple[bytes, bytes]:

        # Key pairs
        indentity_pair = self._bundle.get_indentity()
        prekey_pair = self._bundle.get_prekey()
        onetime_prekey_pair = self._bundle.get_onetime_prekey(opk_id)
        
        SK, message_bytes = OmemoCrypto.accept_init_message(
                encrypted_message=encrypted_message,
                indentity_pair=indentity_pair,
                prekey_pair=prekey_pair,
                onetime_prekey_pair=onetime_prekey_pair,
                indentity_key=indentity_key,
                ephemeral_key=ephemeral_key,
        )

        # Seperate nonces and messsage
        session_nonces = message_bytes[:24]
        message = message_bytes[24:]

        SK_SEND, SK_RECV = OmemoCrypto.split_secret_key(SK)

        self._storage.add_device(jid, device)
        self._storage.add_session(
                jid=jid, 
                device=device, 
                receive_secret_key=b64(SK_RECV), 
                send_secret_key=b64(SK_SEND),
                receive_nonce=b64(session_nonces[12:]),
                send_nonce=b64(session_nonces[:12]), 
        )

        return message

    def send_message(self, jid: str, device: int, message_bytes: bytes) -> Tuple[bytes, bytes, bytes]:
        session = self._storage.get_session(jid, device)
        
        send_nonce = os.urandom(12)
        message = send_nonce + message_bytes

        next_ck, wrapped, payload = OmemoCrypto.send_message(
                ub64(session.send_secret_key), 
                ub64(session.send_nonce), 
                message
        )

        self._storage.update_send_secret(jid, device, b64(next_ck))
        self._storage.set_send_nonce(jid, device, send_nonce)
        return wrapped, payload

    def receive_message(self, jid: str, device: int, wrapped_message_key: bytes, payload: bytes) -> Tuple[bytes, bytes, bytes]:
        session = self._storage.get_session(jid, device)

        next_ck, message_bytes = OmemoCrypto.receive_message(
                ub64(session.receive_secret_key), 
                ub64(session.receive_nonce), 
                wrapped_message_key, 
                payload
        )

        receive_nonce = message_bytes[:12]
        message = message_bytes[12:]

        self._storage.update_receive_secret(jid, device, b64(next_ck))
        self._storage.set_receive_nonce(jid, device, b64(receive_nonce))
        return message

    def close_storage(self):
        pass



import unittest

from osmomemo import XKeyPair
from osmomemo import EdKeyPair
from osmomemo import OmemoBundle

class TestOmemo(unittest.TestCase):
    def test_fingerprint(self):
        bundle = OmemoBundle(
            90909090,
            EdKeyPair.generate(),
            XKeyPair.generate(),
            {
                "0": XKeyPair.generate(),
                "1": XKeyPair.generate(),
                "2": XKeyPair.generate(),
            }
        )

        # Bob
        indentity_fingerprint = bundle.get_indentity_hex_fingerprint()
        b64_public_indentity = bundle.get_indentity().get_base64_public_key()

        # Alice
        public_indentity = EdKeyPair.base64_to_public_key(b64_public_indentity)
        public_indentity_bytes = EdKeyPair.public_key_to_bytes(public_indentity)

        fingerprint = OmemoBundle.public_key_to_hex_fingerprint(public_indentity_bytes)

        self.assertEqual(indentity_fingerprint, fingerprint)






       




from setuptools import setup, find_packages

setup(
    name="osmomemo",
    version="1.1.2",

    author="osmiumnet",
    description="Python omemo package",

    packages=find_packages(),
    install_requires=[
        "cryptography>=45.0.6",
        "pynacl>=1.5.0",
        "sqlalchemy>=2.0.43",
    ],
    python_requires=">=3.10",
)



import os
import json

from typing import Tuple

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey  

from .key import XKeyPair, EdKeyPair 
from .bundle import OmemoBundle

class OmemoCryptography:
    @staticmethod
    def create_init_message(
                message_bytes: bytes,
                indentity_pair: EdKeyPair,
                indentity_key: Ed25519PublicKey,
                signed_prekey: X25519PublicKey,
                prekey_signature: bytes,
                onetime_prekey: X25519PublicKey,
            ) -> Tuple[bytes, bytes, bytes]:

        # Verify signed PreKey signature
        EdKeyPair.verify_public_key(
                verifier=indentity_key,
                public_key=signed_prekey,
                signature=prekey_signature
        )

        # We are the initiator, so we must generate a ephemeral key
        ephemeral_pair = XKeyPair.generate()

        # Private keys
        ik = indentity_pair.get_x_private_key()
        ek = ephemeral_pair.get_private_key()  

        # Make Elliptic Curve Diffie-Hellman parts
        DH1 = ik.exchange(signed_prekey)
        DH2 = ek.exchange(EdKeyPair.public_ed_to_x_key(indentity_key))
        DH3 = ek.exchange(signed_prekey)
        DH4 = ek.exchange(onetime_prekey)

        # Calculate Secret Key
        SK = OmemoCryptography._hkdf_derive(DH1 + DH2 + DH3 + DH4) 

        # Delete ephemeral
        ek_pub = ephemeral_pair.get_public_key()
        del ephemeral_pair  
        del ek

        # Derive an AEAD key and encrypt the initial payload
        aead_key = OmemoCryptography._hkdf_derive(SK)
        aesgcm = AESGCM(aead_key)
        nonce = os.urandom(12)
        ct = aesgcm.encrypt(nonce, message_bytes, None)
        encrypted_message = nonce + ct

        return SK, ek_pub, encrypted_message 

    @staticmethod
    def accept_init_message(
                encrypted_message: bytes,
                indentity_pair: EdKeyPair,
                prekey_pair: XKeyPair,
                onetime_prekey_pair: XKeyPair,
                indentity_key: Ed25519PublicKey,
                ephemeral_key: X25519PublicKey,
            ) -> Tuple[bytes, bytes]:
        
        # Private keys
        ik = indentity_pair.get_x_private_key()
        spk = prekey_pair.get_private_key()
        opk = onetime_prekey_pair.get_private_key()

        # Make Elliptic Curve Diffie-Hellman parts
        DH1 = spk.exchange(EdKeyPair.public_ed_to_x_key(indentity_key))
        DH2 = ik.exchange(ephemeral_key)
        DH3 = spk.exchange(ephemeral_key)
        DH4 = opk.exchange(ephemeral_key)

        # Calculate Secret Key
        SK = OmemoCryptography._hkdf_derive(DH1 + DH2 + DH3 + DH4) 

        # Derive an AEAD key and encrypt the initial payload
        aead_key = OmemoCryptography._hkdf_derive(SK)
        aesgcm = AESGCM(aead_key)
        nonce = encrypted_message[:12]; 
        ct = encrypted_message[12:]
        message_bytes = aesgcm.decrypt(nonce, ct, None)

        return SK, message_bytes

    @staticmethod
    def split_secret_key(secret_key: bytes) -> Tuple[bytes, bytes]:
        two_cks = OmemoCryptography._hkdf_derive(secret_key, length=64)
        return two_cks[32:], two_cks[:32]

    @staticmethod
    def send_message(chain_key: bytes, nonce: bytes, message_bytes: bytes) -> Tuple[bytes, bytes, bytes]:
        msg_key, wrap_key, next_ck = OmemoCryptography._derive_message_and_wrap(chain_key, nonce)
        wrapped = OmemoCryptography._wrap_message_key(wrap_key, msg_key)
        payload = OmemoCryptography._encrypt_payload_with_msgkey(msg_key, message_bytes)
        return next_ck, wrapped, payload

    @staticmethod
    def receive_message(chain_key: bytes, nonce: bytes, wrapped_message_key: bytes, payload: bytes) -> Tuple[bytes, bytes, bytes]:
        _, wrap_key, next_ck = OmemoCryptography._derive_message_and_wrap(chain_key, nonce)
        message_key = OmemoCryptography._unwrap_message_key(wrap_key, wrapped_message_key)
        message = OmemoCryptography._decrypt_payload_with_msgkey(message_key, payload)
        return next_ck, message



    @staticmethod
    def _hkdf_derive(kbs, info=b"OMEMO X3DH", length=32, salt=None) -> bytes:
        hk = HKDF(algorithm=hashes.SHA256(), info=info, length=length, salt=salt)
        return hk.derive(kbs)

    @staticmethod
    def _derive_message_and_wrap(ck: bytes, nonce: bytes) -> Tuple[bytes, bytes, bytes]:
        msg_key = OmemoCryptography._hkdf_derive(ck, info=b"msg|" + nonce, length=32)
        wrap_key = OmemoCryptography._hkdf_derive(ck, info=b"wrap", length=32)
        new_ck = OmemoCryptography._hkdf_derive(ck, info=b"ck", length=32)
        return msg_key, wrap_key, new_ck

    @staticmethod
    def _wrap_message_key(wrap_key: bytes, message_key: bytes) -> bytes:
        aes = AESGCM(wrap_key)
        nonce = os.urandom(12)
        ct = aes.encrypt(nonce, message_key, None)
        return nonce + ct

    @staticmethod
    def _unwrap_message_key(wrap_key: bytes, message_key: bytes) -> bytes:
        nonce, ct = message_key[:12], message_key[12:]
        aes = AESGCM(wrap_key)
        return aes.decrypt(nonce, ct, None)

    @staticmethod
    def _encrypt_payload_with_msgkey(message_key: bytes, message_bytes: bytes) -> bytes:
        aes = AESGCM(message_key)
        nonce = os.urandom(12)
        ct = aes.encrypt(nonce, message_bytes, None)
        return nonce + ct

    @staticmethod
    def _decrypt_payload_with_msgkey(message_key, payload) -> bytes:
        nonce, ct = payload[:12], payload[12:]
        aes = AESGCM(message_key)
        return aes.decrypt(nonce, ct, None)



